@@ -124,8 +124,7 @@ def inline_image(img_file, width):
 
 
 def terminal_image(img_file):
-    # Render by timg is better than old chafa.
-    # it is also centered
+    # This code will try chafa first and, if it fails, try timg
     cmds = []
     if _HAS_CHAFA:
         cmds.append("chafa -C on -d 0 --bg white -w 1")
@@ -146,6 +145,8 @@ def terminal_image(img_file):
 class AbstractRenderer:
     def __init__(self, content, url, center=True,**kwargs):
         self.url = url
+        #base url is used to construct relative urls (see <base> in html)
+        self.base = None
         self.body = str(content)
         # there’s one rendered text and one links table per mode
         self.rendered_text = {}
@@ -366,7 +367,7 @@ class AbstractRenderer:
 
         # Beware, blocks are not wrapped nor indented and left untouched!
         # They are mostly useful for pictures and preformatted text.
-        def add_block(self, intext, theme=None):
+        def add_block(self, intext, theme=None, preformat_wrap=False):
             # If necessary, we add the title before a block
             self._title_first()
             # we don’t want to indent blocks
@@ -374,16 +375,17 @@ class AbstractRenderer:
             self._disable_indents()
             # we have to apply the theme for every line in the intext
             # applying theme to preformatted is controversial as it could change it
+            # We wrap preformatted text if requested or if it is set in the option
             if "preformat_wrap" in self.options:
-                preformat_wrap = self.options["preformat_wrap"]
+                preformwrap = preformat_wrap or self.options["preformat_wrap"]
             else:
-                preformat_wrap = False
+                preformwrap = preformat_wrap
             if theme:
                 block = ""
                 lines = intext.split("\n")
                 for l in lines:
                     self.open_theme(theme)
-                    if preformat_wrap:
+                    if preformwrap:
                         self.add_text(l)
                     else:
                         self.last_line += self.current_indent + l
@@ -455,9 +457,9 @@ class AbstractRenderer:
 
     def get_link(self, nb):
         links = self.get_links()
-        if len(links) < nb:
-            print("Index too high! No link %s for %s" % (nb, self.url))
-            return None
+        if nb not in range(1, len(links)+1):
+            print("%s is not a valid link for %s" % (nb, self.url))
+            return 0
         else:
             return links[nb - 1]
 
@@ -1113,7 +1115,6 @@ class ImageRenderer(AbstractRenderer):
             terminal_image(self.body)
             return True
 
-
 class HtmlRenderer(AbstractRenderer):
     def get_mime(self):
         return "text/html"
@@ -1158,6 +1159,18 @@ class HtmlRenderer(AbstractRenderer):
         else:
             return ""
 
+    def get_base_url(self):
+        if not self.base :
+            if self.body:
+                soup = BeautifulSoup(self.body, "html.parser")
+                if soup.base :
+                    base = soup.base.get("href")
+                    self.base = urllib.parse.urljoin(self.url,base)
+                else:
+                    self.base = self.url
+            else:
+                self.base = self.url
+        return self.base
     # Our own HTML engine (crazy, isn’t it?)
     # Return [rendered_body, list_of_links]
     # mode is either links_only, readable or full
@@ -1180,9 +1193,9 @@ class HtmlRenderer(AbstractRenderer):
         # CR are not meaniningful. Except that, somethimes, they should be interpreted as spaces.
         # HTML is real crap. At least the one people are generating.
 
-        def render_image(src, width=40, mode=None):
+        def render_image(src, width=None, mode=None):
             ansi_img = ""
-            imgurl, imgdata = looks_like_base64(src, self.url)
+            imgurl, imgdata = looks_like_base64(src, self.get_base_url())
             if (
                 _RENDER_IMAGE
                 and mode not in ["full_links_only", "links_only"]
@@ -1198,9 +1211,11 @@ class HtmlRenderer(AbstractRenderer):
                             cached.close()
                     if netcache.is_cache_valid(img):
                         renderer = ImageRenderer(img, imgurl)
-                        # Image are 40px wide except if terminal is smaller
-                        if width > 40:
-                            size = 40
+                        # Image width is set in the option to 40 by default
+                        # it cannot be bigger than the width of the text
+                        if "images_size" in self.options.keys() and width and \
+                                            width > self.options["images_size"] :
+                            size = self.options["images_size"]
                         else:
                             size = width
                         ansi_img = "\n" + renderer.get_body(width=size, mode="inline")
@@ -1215,6 +1230,8 @@ class HtmlRenderer(AbstractRenderer):
             # never start with a "\n"
             # string = string.lstrip("\n")
             string = string.replace("\r", "").replace("\n", " ").replace("\t", " ")
+            # remove soft hyphens
+            string = string.replace("\u00ad","")
             endspace = string.endswith(" ") or string.endswith("\xa0")
             startspace = string.startswith(" ") or string.startswith("\xa0")
             toreturn = string.replace("\n", " ").replace("\t", " ").strip()
@@ -1268,12 +1285,12 @@ class HtmlRenderer(AbstractRenderer):
                     # r.close_all()
                 r.close_all()
                 r.newparagraph()
-            elif element.name in ["code", "tt"]:
+            elif element.name in ["code", "tt","abbr"]:
                 for child in element.children:
                     recursive_render(child, indent=indent, preformatted=True)
             elif element.name in ["pre"]:
                 r.newparagraph()
-                r.add_block(element.text, theme="preformatted")
+                r.add_block(element.text,theme="preformatted",preformat_wrap=True)
                 r.newparagraph(force=True)
             elif element.name in ["li"]:
                 r.startindent(" • ", sub="   ")
@@ -1342,7 +1359,7 @@ class HtmlRenderer(AbstractRenderer):
                 if src:
                     if mode not in self.images:
                         self.images[mode] = []
-                    abs_url, data = looks_like_base64(src, self.url)
+                    abs_url, data = looks_like_base64(src, self.get_base_url())
                     # if abs_url is None, it means we don’t support
                     # the image (such as svg+xml). So we hide it.
                     if abs_url:
@@ -1376,9 +1393,9 @@ class HtmlRenderer(AbstractRenderer):
                 if poster:
                     if mode not in self.images:
                         self.images[mode] = []
-                    poster_url, d = looks_like_base64(poster, self.url)
+                    poster_url, d = looks_like_base64(poster, self.get_base_url())
                     if poster_url:
-                        vid_url, d2 = looks_like_base64(src, self.url)
+                        vid_url, d2 = looks_like_base64(src, self.get_base_url())
                         self.images[mode].append(poster_url)
                         r.add_block(ansi_img)
                         r.open_theme("image_link")
@@ -1392,7 +1409,7 @@ class HtmlRenderer(AbstractRenderer):
                         r.close_theme("image_link")
                         r.newline()
                 elif src:
-                    vid_url, d = looks_like_base64(src, self.url)
+                    vid_url, d = looks_like_base64(src, self.get_base_url())
                     links.append(vid_url + " " + text)
                     link_id = " [%s]" % (len(links) + startlinks)
                     r.open_theme("image_link")
@@ -1403,6 +1420,10 @@ class HtmlRenderer(AbstractRenderer):
 
             elif element.name == "br":
                 r.newline()
+                #weirdly, it seems that BS4 sometimes parse elements after <br>
+                #as children of <br>
+                for child in element.children:
+                    recursive_render(child, indent=indent)
             elif (
                 element.name not in ["script", "style", "template"]
                 and type(element) is not Comment
@@ -1422,6 +1443,7 @@ class HtmlRenderer(AbstractRenderer):
 
         # the real render_html hearth
         # note to vjousse: use self.options["ftr_site_config"] here
+        # Use self.url to mach against an url
         if mode in ["full", "full_links_only"]:
             summary = body
         elif _HAS_READABILITY:
@@ -1441,12 +1463,80 @@ class HtmlRenderer(AbstractRenderer):
                 recursive_render(soup)
         return r.get_final(), links
 
+## Now the custom renderers
+class XkcdRenderer(HtmlRenderer):
+    def printgemtext(self,source):
+        if source:
+            gemtext_renderer = GemtextRenderer("",self.url)
+            final,links = gemtext_renderer.render(source)
+            print(final)
+
+    def has_direct_display(self):
+        return _RENDER_IMAGE
+
+    #Custom renderer should return false when they can’t handle a specific content
+    #This allow fallback to normal html renderer
+    def is_valid(self):
+        path = urllib.parse.urlparse(self.url).path
+        #We strip the leading "/" to avoid empty elements in splitted
+        splitted = path.strip("/").split("/")
+        # Custom renderer only works for pure comics which have alphanumeric paths 
+        return splitted[0].isalnum()
+
+    def get_images(self, mode="readable"):
+        img_url,img_path,alttext,title = self.xkcd_extract()
+        return [img_url]
+
+    def get_links(self,mode="readable"):
+        img_url,img_path,alttext,title = self.xkcd_extract()
+        links = super().get_links(mode=mode)
+        if img_url not in links:
+            links.append(img_url)
+        return links
+
+    #return [image_url,image_path, image_alt_text,image_title]
+    def xkcd_extract(self):
+        soup = BeautifulSoup(self.body, "html.parser")
+        comic_div = soup.find("div",{"id":"comic"})
+        if comic_div:
+            img_element = comic_div.find("img")
+            src=img_element.get("src")
+            if src.startswith("//"):
+                scheme = urllib.parse.urlparse(self.url).scheme
+                img_url = scheme + ":" + src
+            else:
+                img_url = src
+            img_path = netcache.get_cache_path(img_url)
+            alttext=img_element.get("title")
+            title=img_element.get("alt")
+            return img_url,img_path,alttext,title
+        else:
+            return None,None,None,None
+
+    def display(self, mode=None, directdisplay=False):
+        wtitle = self.get_formatted_title()
+        if not directdisplay:
+            body = wtitle + "\n" + self.get_body(mode=mode)
+            return body
+        else:
+            print(self._window_title(wtitle))
+            self.printgemtext("# "+self.get_title())
+            img_url,img_path,alttext,title = self.xkcd_extract()
+            #now displaying
+            if img_path and netcache.is_cache_valid(img_url):
+                terminal_image(img_path)
+            else:
+                self.printgemtext("\n> missing picture, please reload\n")
+            self.printgemtext(alttext)
+            return True
+
 
 # Mapping mimetypes with renderers
 # (any content with a mimetype text/* not listed here will be rendered with as GemText)
 _FORMAT_RENDERERS = {
     "text/gemini": GemtextRenderer,
     "text/html": HtmlRenderer,
+    "application/xhtml+xml": HtmlRenderer,
     "text/xml": FeedRenderer,
     "text/plain": PlaintextRenderer,
     "application/xml": FeedRenderer,
@@ -1457,8 +1547,14 @@ _FORMAT_RENDERERS = {
     "application/javascript": HtmlRenderer,
     "application/json": HtmlRenderer,
     "text/empty": EmptyRenderer,
+    "message/news": GemtextRenderer,
+    "message/rfc822": GemtextRenderer,
 }
 
+_CUSTOM_RENDERERS = {
+        "xkcd.com": XkcdRenderer,
+        }
+
 
 def get_mime(path, url=None):
     # Beware, this one is really a shaddy ad-hoc function
@@ -1478,12 +1574,17 @@ def get_mime(path, url=None):
             itemtype = "1"
             path = ""
         if itemtype == "0":
-            mime = "text/gemini"
+            if path.endswith(".xml"):
+                mime = "application/xml"
+            else:
+                mime = "text/gemini"
         elif itemtype == "1":
             mime = "text/gopher"
         elif itemtype == "h":
             mime = "text/html"
-        elif itemtype in ("9", "g", "I", "s", ";"):
+        elif itemtype in ("g", "I"):
+            mime = mimetypes.guess_type(path)[0]
+        elif itemtype in ("9", "s", ";"):
             mime = "binary"
         else:
             mime = "text/gopher"
@@ -1502,6 +1603,19 @@ def get_mime(path, url=None):
         # because the FeedRendered fallback to HtmlRenderer
         if mime2 and mime != mime2 and "html" in mime and "xml" in mime2:
             mime = "text/xml"
+        # We will also check if the first line starts with <rss
+        elif not mime2 and "html" in mime:
+            if path and os.path.exists(path):
+                firstline = None
+                with open(path) as f:
+                    firstline = f.readline()
+                    f.close()
+                if firstline.startswith("<rss"):
+                    mime = "text/xml"
+                else:
+                    mime = "text/html"
+            else:
+                mime = "text/html"
         # If it’s a xml file, consider it as such, regardless of what file thinks
         elif path.endswith(".xml"):
             mime = "text/xml"
@@ -1556,29 +1670,38 @@ def set_renderer(content, url, mime, theme=None,**kwargs):
         if theme:
             renderer.set_theme(theme)
         return renderer
-    mime_to_use = []
-    for m in _FORMAT_RENDERERS:
-        if fnmatch.fnmatch(mime, m):
-            mime_to_use.append(m)
-    if len(mime_to_use) > 0:
-        current_mime = mime_to_use[0]
-        func = _FORMAT_RENDERERS[current_mime]
-        if current_mime.startswith("text"):
+    # First, we check for a custom renderer based on url
+    if url:
+        netloc = urllib.parse.urlparse(url).netloc
+        is_http = urllib.parse.urlparse(url).scheme in ["http","https"]
+        #custom renderer are only for http/https, right?
+        if is_http and netloc and netloc in _CUSTOM_RENDERERS:
+            func = _CUSTOM_RENDERERS[netloc]
             renderer = func(content, url,**kwargs)
-            # We double check if the renderer is correct.
-            # If not, we fallback to html
-            # (this is currently only for XHTML, often being
-            # mislabelled as xml thus RSS feeds)
-            if not renderer.is_valid():
-                func = _FORMAT_RENDERERS["text/html"]
-                # print("Set (fallback)RENDERER to html instead of %s"%mime)
+    if not renderer or not renderer.is_valid():
+        mime_to_use = []
+        for m in _FORMAT_RENDERERS:
+            if fnmatch.fnmatch(mime, m):
+                mime_to_use.append(m)
+        while len(mime_to_use) > 0 and (not renderer or not renderer.is_valid()):
+            current_mime = mime_to_use.pop(0)
+            func = _FORMAT_RENDERERS[current_mime]
+            if current_mime.startswith("text"):
                 renderer = func(content, url,**kwargs)
-        else:
-            # TODO: check this code and then remove one if.
-            # we don’t parse text, we give the file to the renderer
-            renderer = func(content, url,**kwargs)
-            if not renderer.is_valid():
-                renderer = None
+                # We double check if the renderer is correct.
+                # If not, we fallback to html
+                # (this is currently only for XHTML, often being
+                # mislabelled as xml thus RSS feeds)
+                if not renderer.is_valid():
+                    func = _FORMAT_RENDERERS["text/html"]
+                    # print("Set (fallback)RENDERER to html instead of %s"%mime)
+                    renderer = func(content, url,**kwargs)
+            else:
+                # TODO: check this code and then remove one if.
+                # we don’t parse text, we give the file to the renderer
+                renderer = func(content, url,**kwargs)
+                if not renderer.is_valid():
+                    renderer = None
     #We have not found a renderer. Use a fake one.
     if not renderer:
         renderer = FakeRenderer("",url,**kwargs)
@@ -1588,6 +1711,7 @@ def set_renderer(content, url, mime, theme=None,**kwargs):
     return renderer
 
 
+# This function should be removed and replaced by a set_renderer()/r.display()
 def render(input, path=None, format="auto", mime=None, url=None, mode=None):
     if not url:
         url = ""
